{-# LANGUAGE DerivingStrategies  #-}
{-# LANGUAGE OverloadedStrings   #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications    #-}
module Test.Core.Text.Corpus.Query (tests) where

import Data.BoolExpr
import Data.Conduit
import Data.Maybe
import Data.String
import Gargantext.Core (Lang(..), toISO639)
import Gargantext.Core.Text.Corpus.Query
import Gargantext.Core.Types
import Gargantext.Database.Admin.Types.Hyperdata (HyperdataDocument(..))
import Prelude
import System.Environment
import qualified Data.Text as T
import qualified Gargantext.Core.Text.Corpus.API.Arxiv as Arxiv
import qualified Gargantext.Core.Text.Corpus.API.Pubmed as Pubmed
import qualified Gargantext.Core.Text.Corpus.API.Isidore as Isidore
import qualified Gargantext.Core.Text.Corpus.API.Istex   as Istex
import qualified Gargantext.Core.Text.Corpus.API.Hal     as Hal
import qualified Network.Api.Arxiv as Arxiv

import Test.Tasty
import Test.Tasty.HUnit
import Test.Tasty.QuickCheck hiding (Positive, Negative)

newtype PubmedApiKey
  = PubmedApiKey { _PubmedApiKey :: T.Text }
  deriving stock (Show, Eq)
  deriving newtype IsString

pubmedSettings :: IO (Maybe PubmedApiKey)
pubmedSettings = fmap fromString <$> lookupEnv "PUBMED_API_KEY"

tests :: TestTree
tests = withResource pubmedSettings (const (pure ())) $ \getPubmedKey ->
  testGroup "Boolean Query Engine" [
    testProperty "Parses 'A OR B'" testParse01
  , testProperty "Parses 'A AND B'" testParse02
  , testProperty "Parses '-A'" testParse03
  , testProperty "Parses 'NOT A'" testParse03_01
  , testProperty "Parses 'A -B'" testParse04
  , testProperty "Parses 'A NOT -B'" testParse04_01
  , testProperty "Parses 'A AND B -C'   (left associative)" testParse05
  , testProperty "Parses 'A AND (B -C)' (right associative)" testParse05_01
  , testProperty "Parses (A OR B OR NOT C) AND (D OR E OR F) -(G OR H OR I)" testParse06
  , testProperty "It supports '\"Haskell\" AND \"Idris\"'" testParse07
  , testProperty "It supports 'Haskell AND Idris'" testParse07_01
  , testProperty "It supports 'Raphael'" testParse07_02
  , testProperty "It supports 'Niki', 'Ajeje' and 'Orf'" testParse07_03
  , testCase     "Parses words into a single constant" testWordsIntoConst
  , testGroup "Arxiv expression converter" [
      testCase "It supports 'A AND B'"     testArxiv01_01
    , testCase "It supports '\"Haskell\" AND \"Agda\"'" testArxiv01_02
    , testCase "It supports 'A OR B'"        testArxiv02
    , testCase "It supports 'A AND NOT B'"   testArxiv03_01
    , testCase "It supports 'A AND -B'"      testArxiv03_02
    , testCase "It supports 'A AND -B'"      testArxiv03_02
    , testCase "It supports 'A AND NOT (NOT B)'"  testArxiv04_01
    , testCase "It supports 'A AND NOT (NOT (NOT B))'"  testArxiv04_02
    , testCase "It supports 'A OR NOT B'"  testArxiv05
    , testCase "It supports '-A'"  testArxiv06
    ]
  , testGroup "PUBMED expression converter" [
      testCase "It supports 'A'" testPubMed01
    , testCase "It supports '-A'" testPubMed02_01
    , testCase "It supports 'NOT A'" testPubMed02_02
    , testCase "It supports 'NOT (NOT A)'" testPubMed02_03
    , testCase "It supports '\"Haskell\" AND \"Idris\"'" testPubMed03
    , testCase "It supports 'A OR B'" testPubMed04
    ]
  , testGroup "Isidore expression converter" [
      testCase "It supports 'A'" testIsidore01
    , testCase "It supports '-A'" testIsidore02_01
    , testCase "It supports 'NOT A'" testIsidore02_02
    , testCase "It supports 'NOT (NOT A)'" testIsidore02_03
    , testCase "It supports '\"Haskell\" AND \"Idris\"'" testIsidore03
    , testCase "It supports 'A OR B'" testIsidore04
    ]
  , testGroup "Istex expression converter" [
      testCase "It supports 'A'" testIstex01
    , testCase "It supports '-A'" testIstex02_01
    , testCase "It supports 'NOT A'" testIstex02_02
    , testCase "It supports 'NOT (NOT A)'" testIstex02_03
    , testCase "It supports '\"Haskell\" AND \"Idris\"'" testIstex03
    , testCase "It supports 'A OR B'" testIstex04
    ]
  , testGroup "HAL expression converter" [
      testCase "It supports 'A'" testHal01
    , testCase "It supports '-A'" testHal02_01
    , testCase "It supports 'NOT A'" testHal02_02
    , testCase "It supports 'NOT (NOT A)'" testHal02_03
    , testCase "It supports '\"Haskell\" AND \"Idris\"'" testHal03
    , testCase "It supports 'A OR B'" testHal04
    , testCase "It supports 'Paris AND France AND history NOT (Texas AND history)'" testHal05
    ]
  , testGroup "PUBMED real queries (skipped if PUBMED_API_KEY env var not set)" [
      testCase "It searches for \"Covid\"" (testPubMedCovid_01 getPubmedKey)
    , testCase "It searches for \"Covid\" AND \"Alzheimer\"" (testPubMedCovid_02 getPubmedKey)
    ]
  -- We skip the Arxiv tests if the PUBMED_API_KEY is not set just for conveniency, to have
  -- only a single flow-control mechanism.
  , testGroup "ARXIV real queries (skipped if PUBMED_API_KEY env var not set)" [
      testCase "It searches for \"Haskell\"" (testArxivRealWorld_01 getPubmedKey)
    , testCase "It searches for \"Haskell\" AND \"Agda\"" (testArxivRealWorld_02 getPubmedKey)
    ]
  -- .. ditto for Isidore
  , testGroup "Isidore real queries (skipped if PUBMED_API_KEY env var not set)" [
      testCase "It searches for \"chicken pox\"" (testIsidoreRealWorld_01 getPubmedKey)
    , testCase "It searches for \"Dante\" AND \"Petrarca\"" (testIsidoreRealWorld_02 getPubmedKey)
    ]
  -- .. ditto for Istex
  , testGroup "Istex real queries (skipped if PUBMED_API_KEY env var not set)" [
      testCase "It searches for \"brain\"" (testIstexRealWorld_01 getPubmedKey)
    , testCase "It searches for \"brain\" AND NOT \"neural\"" (testIstexRealWorld_02 getPubmedKey)
    ]
  -- .. ditto for HAL
  , testGroup "HAL real queries (skipped if PUBMED_API_KEY env var not set)" [
      testCase "It searches for \"Haskell\"" (testHalRealWorld_01 getPubmedKey)
    , testCase "It searches for \"Haskell\" AND NOT \"Agda\"" (testHalRealWorld_02 getPubmedKey)
    ]
  ]

-- | Checks that the 'RawQuery' can be translated into the expected 'BoolExpr' form,
-- by also checking that both renders back to the initial 'RawQuery'.
translatesInto :: RawQuery -> BoolExpr Term -> Property
(translatesInto) raw boolExpr =
  let parsed   = parseQuery raw
      expected = Right (unsafeMkQuery boolExpr)
  in counterexample (show parsed <> " != " <> show expected) $
       (renderQuery <$> parsed) === (renderQuery <$> expected)

testParse01 :: Property
testParse01 = "A OR B" `translatesInto` (BConst (Positive "A") `BOr` BConst (Positive "B"))

testParse02 :: Property
testParse02 = "A AND B" `translatesInto` (BConst (Positive "A") `BAnd` BConst (Positive "B"))

testParse03 :: Property
testParse03 = "-A" `translatesInto` (BConst (Negative "A"))

testParse03_01 :: Property
testParse03_01 = "NOT A" `translatesInto` (BConst (Negative "A"))

testParse04 :: Property
testParse04 = "A -B" `translatesInto` (BConst (Positive "A") `BAnd` BConst (Negative "B"))

-- Both 'A -B' and 'A AND -B' desugars into the same form.
testParse04_01 :: Property
testParse04_01 = "A AND -B" `translatesInto` (BConst (Positive "A") `BAnd` BConst (Negative "B"))

testParse05 :: Property
testParse05 = "A AND B -C" `translatesInto` ((BConst (Positive "A") `BAnd` BConst (Positive "B")) `BAnd` BConst (Negative "C"))

testParse05_01 :: Property
testParse05_01 =
  "A AND (B -C)" `translatesInto` (BConst (Positive "A") `BAnd` (BConst (Positive "B") `BAnd` BConst (Negative "C")))

testParse06 :: Property
testParse06 =
  translatesInto "(A OR B OR NOT C) AND (D OR E OR F) -(G OR H OR I)"
    (
      (
        ((BConst (Positive  "A") `BOr` (BConst (Positive  "B"))) `BOr` (BConst  (Negative "C")))
        `BAnd`
        ((BConst (Positive "D") `BOr` (BConst (Positive "E"))) `BOr` (BConst (Positive "F")))
      )
      `BAnd` BNot (
        ((BConst (Positive "G") `BOr` (BConst (Positive "H"))) `BOr` (BConst (Positive "I")))
      )
    )

testParse07 :: Property
testParse07 =
  translatesInto "\"Haskell\" AND \"Agda\""
    ((BConst (Positive  "Haskell") `BAnd` (BConst (Positive  "Agda"))))

testParse07_01 :: Property
testParse07_01 =
  translatesInto "Haskell AND Agda"
    ((BConst (Positive  "Haskell") `BAnd` (BConst (Positive  "Agda"))))

testParse07_02 :: Property
testParse07_02 =
  translatesInto "Raphael"
    ((BConst (Positive  "Raphael")))

testParse07_03 :: Property
testParse07_03 =
  translatesInto "Niki" ((BConst (Positive "Niki"))) .&&.
  translatesInto "Ajeje" ((BConst (Positive "Ajeje"))) .&&.
  translatesInto "Orf" ((BConst (Positive "Orf")))

testWordsIntoConst :: Assertion
testWordsIntoConst =
  let (expected :: BoolExpr Term) =
        fromCNF (boolTreeToCNF @Term $ (BConst (Positive "The Art of Computer Programming") `BAnd` (BConst (Positive "Conceptual Mathematics"))))
  in case parseQuery "\"The Art of Computer Programming\" AND \"Conceptual Mathematics\"" of
    Left err
      -> assertBool err False
    Right x
      -> fromCNF (getQuery x) @?= expected

withValidQuery :: RawQuery -> (Query -> Assertion) -> Assertion
withValidQuery rawQuery onValidParse = do
  case parseQuery rawQuery of
    Left err -> assertBool err False
    Right x  -> onValidParse x


testArxiv01_01 :: Assertion
testArxiv01_01 = withValidQuery "A AND B" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (Arxiv.And (Arxiv.Exp $ Arxiv.Abs ["A"]) ((Arxiv.Exp $ Arxiv.Abs ["B"]))))

testArxiv01_02 :: Assertion
testArxiv01_02 = withValidQuery "\"Haskell\" AND \"Agda\"" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (Arxiv.And (Arxiv.Exp $ Arxiv.Abs ["Haskell"]) ((Arxiv.Exp $ Arxiv.Abs ["Agda"]))))

testArxiv02 :: Assertion
testArxiv02 = withValidQuery "A OR B" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (Arxiv.Or (Arxiv.Exp $ Arxiv.Abs ["A"]) ((Arxiv.Exp $ Arxiv.Abs ["B"]))))

testArxiv03_01 :: Assertion
testArxiv03_01 = withValidQuery "A AND NOT B" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (Arxiv.AndNot (Arxiv.Exp $ Arxiv.Abs ["A"]) ((Arxiv.Exp $ Arxiv.Abs ["B"]))))

testArxiv03_02 :: Assertion
testArxiv03_02 = withValidQuery "A AND -B" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (Arxiv.AndNot (Arxiv.Exp $ Arxiv.Abs ["A"]) ((Arxiv.Exp $ Arxiv.Abs ["B"]))))

-- Double negation get turned into positive.
testArxiv04_01 :: Assertion
testArxiv04_01 = withValidQuery "A AND NOT (NOT B)" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (Arxiv.And (Arxiv.Exp $ Arxiv.Abs ["A"]) ((Arxiv.Exp $ Arxiv.Abs ["B"]))))

testArxiv04_02 :: Assertion
testArxiv04_02 = withValidQuery "A AND NOT (NOT (NOT B))" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (Arxiv.AndNot (Arxiv.Exp $ Arxiv.Abs ["A"]) ((Arxiv.Exp $ Arxiv.Abs ["B"]))))

testArxiv05 :: Assertion
testArxiv05 = withValidQuery "A OR NOT B" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (
                Arxiv.Or (Arxiv.Exp $ Arxiv.Abs ["A"])
                         (Arxiv.AndNot (Arxiv.Exp $ Arxiv.Abs ["B"]) (Arxiv.Exp $ Arxiv.Abs ["B"]))
                )
             )

testArxiv06 :: Assertion
testArxiv06 = withValidQuery "-A" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Arxiv.qExp (Arxiv.convertQuery q) == Just (
                Arxiv.AndNot (Arxiv.Exp $ Arxiv.Abs ["A"]) (Arxiv.Exp $ Arxiv.Abs ["A"])
                )
             )

--
-- PUBMED tests
--

testPubMed01 :: Assertion
testPubMed01 = withValidQuery "A" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Pubmed.getESearch (Pubmed.convertQuery q) == "A")

testPubMed02_01 :: Assertion
testPubMed02_01 = withValidQuery "-A" $ \q -> Pubmed.getESearch (Pubmed.convertQuery q) @?= "NOT+A"

testPubMed02_02 :: Assertion
testPubMed02_02 = withValidQuery "NOT A" $ \q -> Pubmed.getESearch (Pubmed.convertQuery q) @?= "NOT+A"

testPubMed02_03 :: Assertion
testPubMed02_03 = withValidQuery "NOT (NOT A)" $ \q -> Pubmed.getESearch (Pubmed.convertQuery q) @?= "A"

testPubMed03 :: Assertion
testPubMed03 = withValidQuery "\"Haskell\" AND \"Idris\"" $ \q ->
  Pubmed.getESearch (Pubmed.convertQuery q) @?= "Haskell+AND+Idris"

testPubMed04 :: Assertion
testPubMed04 = withValidQuery "A OR B" $ \q ->
  Pubmed.getESearch (Pubmed.convertQuery q) @?= "A+OR+B"

--
-- Isidore tests
--

testIsidore01 :: Assertion
testIsidore01 = withValidQuery "A" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Isidore.getIsidoreQuery (Isidore.convertQuery q) == "A")

testIsidore02_01 :: Assertion
testIsidore02_01 = withValidQuery "-A" $ \q -> Isidore.getIsidoreQuery (Isidore.convertQuery q) @?= "NOT+A"

testIsidore02_02 :: Assertion
testIsidore02_02 = withValidQuery "NOT A" $ \q -> Isidore.getIsidoreQuery (Isidore.convertQuery q) @?= "NOT+A"

testIsidore02_03 :: Assertion
testIsidore02_03 = withValidQuery "NOT (NOT A)" $ \q -> Isidore.getIsidoreQuery (Isidore.convertQuery q) @?= "A"

testIsidore03 :: Assertion
testIsidore03 = withValidQuery "\"Haskell\" AND \"Idris\"" $ \q ->
  Isidore.getIsidoreQuery (Isidore.convertQuery q) @?= "Haskell+AND+Idris"

testIsidore04 :: Assertion
testIsidore04 = withValidQuery "A OR B" $ \q ->
  Isidore.getIsidoreQuery (Isidore.convertQuery q) @?= "A+OR+B"

--
-- Istex tests
--

testIstex01 :: Assertion
testIstex01 = withValidQuery "A" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Istex.getIstexQuery EN (Istex.convertQuery q) == "language:eng+AND+abstract:A")

testIstex02_01 :: Assertion
testIstex02_01 = withValidQuery "-A" $ \q -> Istex.getIstexQuery EN (Istex.convertQuery q) @?= "language:eng+AND+NOT+abstract:A"

testIstex02_02 :: Assertion
testIstex02_02 = withValidQuery "NOT A" $ \q -> Istex.getIstexQuery EN (Istex.convertQuery q) @?= "language:eng+AND+NOT+abstract:A"

testIstex02_03 :: Assertion
testIstex02_03 = withValidQuery "NOT (NOT A)" $ \q -> Istex.getIstexQuery EN (Istex.convertQuery q) @?= "language:eng+AND+abstract:A"

testIstex03 :: Assertion
testIstex03 = withValidQuery "\"Haskell\" AND \"Idris\"" $ \q ->
  Istex.getIstexQuery FR (Istex.convertQuery q) @?= "language:fre+AND+abstract:Haskell+AND+abstract:Idris"

testIstex04 :: Assertion
testIstex04 = withValidQuery "A OR B" $ \q ->
  Istex.getIstexQuery EN (Istex.convertQuery q) @?= "language:eng+AND+abstract:A+OR+abstract:B"

--
-- Hal tests
--

testHal01 :: Assertion
testHal01 = withValidQuery "A" $ \q ->
  assertBool ("Query not converted into expression: " <> show @(BoolExpr Term) (fromCNF $ getQuery q))
             (Hal.getHalQuery (Hal.convertQuery q) == "A")

testHal02_01 :: Assertion
testHal02_01 = withValidQuery "-A" $ \q ->
  Hal.getHalQuery (Hal.convertQuery q) @?= "NOT+A"

testHal02_02 :: Assertion
testHal02_02 = withValidQuery "NOT A" $ \q ->
  Hal.getHalQuery (Hal.convertQuery q) @?= "NOT+A"

testHal02_03 :: Assertion
testHal02_03 = withValidQuery "NOT (NOT A)" $ \q ->
  Hal.getHalQuery (Hal.convertQuery q) @?= "A"

testHal03 :: Assertion
testHal03 = withValidQuery "\"Haskell\" AND \"Idris\"" $ \q ->
  Hal.getHalQuery (Hal.convertQuery q) @?= "(Haskell+AND+Idris)"

testHal04 :: Assertion
testHal04 = withValidQuery "A OR B" $ \q ->
  Hal.getHalQuery (Hal.convertQuery q) @?= "(A+OR+B)"

testHal05 :: Assertion
testHal05 = withValidQuery "Paris AND France AND history NOT (Texas AND history)" $ \q ->
  Hal.getHalQuery (Hal.convertQuery q) @?= "(Paris+AND+(France+AND+(history+AND+(NOT+Texas+OR+NOT+history))))"

--
-- Integration tests against the real services
--

testPubMedCovid_01 :: IO (Maybe PubmedApiKey) -> Assertion
testPubMedCovid_01 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just k  -> withValidQuery "\"Covid\"" $ \query -> do
      res <- Pubmed.get (_PubmedApiKey k) (renderQuery query) (Just 1)
      case res of
        Left err       -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> assertBool ("Found: " <> show (_hd_title x))
                                (maybe False (T.isInfixOf "COVID") (_hd_title x))

testPubMedCovid_02 :: IO (Maybe PubmedApiKey) -> Assertion
testPubMedCovid_02 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just k  -> withValidQuery "\"Covid\" AND \"Alzheimer\"" $ \query -> do
      res <- Pubmed.get (_PubmedApiKey k) (renderQuery query) (Just 1)
      case res of
        Left err       -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> _hd_title x @?= Just "Neurodegenerative and Neurodevelopmental Diseases and the Gut-Brain Axis: The Potential of Therapeutic Targeting of the Microbiome."

testArxivRealWorld_01 :: IO (Maybe PubmedApiKey) -> Assertion
testArxivRealWorld_01 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "\"Haskell\"" $ \query -> do
      res <- Arxiv.get EN query (Just 1)
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> _hd_title x @?= Just "Haskell for OCaml programmers"

testArxivRealWorld_02 :: IO (Maybe PubmedApiKey) -> Assertion
testArxivRealWorld_02 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "\"Haskell\" AND \"Agda\"" $ \query -> do
      res <- Arxiv.get EN query (Just 1)
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> _hd_title x @?= Just "Toward Hole-Driven Development with Liquid Haskell"

--
-- Isidore integration tests
--

testIsidoreRealWorld_01 :: IO (Maybe PubmedApiKey) -> Assertion
testIsidoreRealWorld_01 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "\"Chicken Pox\"" $ \query -> do
      res <- Isidore.get EN (Just 1) query Nothing
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> assertBool ("found: " <> show (_hd_title x))
                                (maybe False (T.isInfixOf "chicken pox") (_hd_title x))

testIsidoreRealWorld_02 :: IO (Maybe PubmedApiKey) -> Assertion
testIsidoreRealWorld_02 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "\"dante\" AND \"petrarca\"" $ \query -> do
      res <- Isidore.get EN (Just 1) query Nothing
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> isJust (_hd_title x) @?= True

--
-- Istex integration tests
--

testIstexRealWorld_01 :: IO (Maybe PubmedApiKey) -> Assertion
testIstexRealWorld_01 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "\"brain\"" $ \query -> do
      res <- Istex.get EN query (Just 1)
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> assertBool ("found: " <> show (_hd_title x))
                                (maybe False (T.isInfixOf "brain") (_hd_title x))

testIstexRealWorld_02 :: IO (Maybe PubmedApiKey) -> Assertion
testIstexRealWorld_02 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "\"brain\" AND NOT \"neural\"" $ \query -> do
      res <- Istex.get EN query (Just 1)
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> isJust (_hd_title x) @?= True

--
-- HAL integration tests
--

testHalRealWorld_01 :: IO (Maybe PubmedApiKey) -> Assertion
testHalRealWorld_01 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "Haskell" $ \query -> do
      res <- Hal.getC (toISO639 EN) query (Just 1)
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> assertBool ("found: " <> show (_hd_title x))
                                (isJust (_hd_title x))

testHalRealWorld_02 :: IO (Maybe PubmedApiKey) -> Assertion
testHalRealWorld_02 getPubmedKey = do
  mb_key <- getPubmedKey
  case mb_key of
    Nothing -> pure ()
    Just _  -> withValidQuery "Haskell AND NOT Agda" $ \query -> do
      res <- Hal.getC (toISO639 EN) query (Just 1)
      case res of
        Left err -> fail (show err)
        Right (_, cnd) -> do
          hyperDocs <- sourceToList cnd
          case hyperDocs of
            []    -> fail "No documents found."
            (x:_) -> isJust (_hd_title x) @?= True
